"""
drugstone.task.tasks

This module implements the class :class:`Tasks` for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

from typing import List
from .task import Task
from .task_result import TaskResult
from .tasks_result import TasksResult


class Tasks:
    """Wraps a list of :class:`Task` objects.

    get_result() -> :class:`TasksResult`:
        Returns a :class:`TasksResult` for the list of tasks.

    get_union() -> :class:`TaskResult`:
        Returns a :class:`TaskResult` with the union of the tasks.

    get_intersection() -> :class:`TaskResult`:
        Returns a :class:`TaskResult` with the intersection of the tasks.
    """

    def __init__(self, tasks: List[Task] = []) -> None:
        self.__tasks = tasks

    def get_result(self) -> TasksResult:
        """Returns a :class:`TasksResult` for the list of tasks."""

        return TasksResult(self.__tasks)

    def get_union(self) -> TaskResult:
        """Returns a :class:`TaskResult` with the union of the tasks."""

        drugs = {}
        genes = {}
        for t in self.__tasks:
            r = t.get_result()
            drugs = {**drugs, **r.get_drugs()}
            for gene, detail in r.get_genes().items():
                if gene not in genes:
                    genes = {**genes, gene: detail}
                else:
                    edges_a = genes[gene]["hasEdgesTo"]
                    edges_b = detail["hasEdgesTo"]
                    new_edges = list(set(edges_a + edges_b))
                    genes[gene]["hasEdgesTo"] = new_edges
        for d in drugs:
            if "score" in d:
                drugs[d].pop("score")
        for g in genes:
            if "score" in g:
                genes[g].pop("score")
        return TaskResult(drugs=drugs, genes=genes)

    def get_intersection(self) -> TaskResult:
        """Returns a :class:`TaskResult` with the intersection of the tasks."""

        drugs = {}
        genes = {}
        first = True
        for t in self.__tasks:
            r = t.get_result()
            if first:
                drugs = r.get_drugs()
                genes = r.get_genes()
                first = False
            else:
                drugs = {d: drugs[d] for d in drugs if d in r.get_drugs()}
                genes_intersection = {}
                for gene, detail in r.get_genes().items():
                    if gene in genes:
                        genes_intersection = {**genes_intersection, gene: detail}
                        edges_a = genes[gene]["hasEdgesTo"]
                        edges_b = detail["hasEdgesTo"]
                        new_edges = list(set(edges_a) & set(edges_b))
                        genes_intersection[gene]["hasEdgesTo"] = new_edges
                genes = genes_intersection.copy()
        for d in drugs:
            if "score" in d:
                drugs[d].pop("score")
        for g in genes:
            if "score" in g:
                genes[g].pop("score")
        return TaskResult(drugs=drugs, genes=genes)



"""
drugstone.scripts.map_nodes_to_internal_ids

This module implements the map_nodes_to_internal_ids function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import requests
import warnings
from .constants.url import Url
from .constants.task_parameter import TaskParameter
from .normalize_task_parameter import normalize_task_parameter


def map_nodes_to_internal_ids(
        nodes: list,
        params: dict = dict({})) -> list:
    """Maps the given nodes to the internal IDs.

    Returns a list of the internal IDs.
    """

    if "identifier" in params:
        if params["identifier"] in TaskParameter.IdentifierValues.IDENTIFIER_VALUES:
            identifier = params["identifier"]
        else:
            identifier = get_default_identifier()
            warnings.warn("The identifier: " + str(params["identifier"])
                          + "is not known to Drugstone!"
                          + " The identifier is changed to "
                          + str(identifier) + "!")
    else:
        identifier = get_default_identifier()

    formatted_nodes = []
    for node in nodes:
        formatted_nodes.append({"id": str(node)})

    data = {
        "nodes": formatted_nodes,
        "identifier": identifier
    }

    # Sends the nodes to the drugstone API
    # and receives an extended list with the internal IDs.
    mapped_nodes = requests.post(
        Url.MAP_NODES,
        verify=False,
        json=data
    ).json()
    return mapped_nodes


def get_default_identifier() -> str:
    parameters = normalize_task_parameter({}, [])
    return parameters["parameters"]["config"]["identifier"]



"""
drugstone.new_task

This module implements the new_task function for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import warnings
from typing import List
from .task.task import Task
from .task.models.drug import Drug
from .task.models.gene import Gene
from .scripts.normalize_task_parameter import normalize_task_parameter
from .scripts.wait_for_task_to_finish import wait_for_task_to_finish
from .scripts.map_nodes_to_internal_ids import map_nodes_to_internal_ids
from .scripts.start_task import start_task
from .scripts.request_task_result import request_task_result
from .scripts.fetch_edges import fetch_edges
from .scripts.add_edges_to_genes import add_edges_to_genes
from .scripts.merge_results import merge_results
from .scripts.normalize_nodes import normalize_nodes
from .license import license


def new_task(
        seeds: list = list([]),
        parameters: dict = dict({}),
        static: bool = False,
        static_result: dict = dict({}),
        static_drugs: List[Drug] = list([]),
        static_genes: List[Gene] = list([])) -> Task:
    """Returns a task.

    Starts a task, according to the user given seeds and parameters.
    Returns a :class:`Task` object, representing the task.

    :param list seeds: List of seed nodes for the task.
    :param dict parameters: (optional) Dictionary of parameters for the task. Defaults to an empty dict {}.
    :param bool static: (optional)
    :param dict static_result: (optional)
    :param List[Drug] static_drugs: (optional)
    :param List[Gene] static_genes: (optional)
    :return: :class:`Task` object
    """

    extended_genes = map_nodes_to_internal_ids(seeds, parameters)

    if 'identifier' not in parameters:
        parameters['identifier'] = 'symbol'

    internal_ids = [n[parameters['identifier']][0] for n in extended_genes if n['drugstoneType'] == 'protein']
    normalized_params = normalize_task_parameter(parameters, internal_ids)
    if parameters['algorithm'] == 'adjacentDrugs':
        from .fetching import getAdjacentDrugs
        result = getAdjacentDrugs(normalized_params, extended_genes)
        return Task(result=result, params=normalized_params, raw_data=result)

    # static task
    if static:
        dataset = normalized_params["parameters"]["ppiDataset"]
        edges = fetch_edges(internal_ids, dataset)
        task_result = add_edges_to_genes(extended_genes, edges, parameters['identifier'])
        genes_drugs = __get_dict_for_genes_and_drugs(static_drugs, static_genes)
        task_result = merge_results(task_result, genes_drugs)
        task_result = merge_results(task_result, static_result)
        return Task(result=task_result, params=normalized_params)

    # no seeds and no static
    if not static and not seeds:
        warnings.warn("Something went wrong! "
                      + "Maybe you forgot to pass seed genes, "
                      + "or forgot to set static to True.")
        return Task()

    # dynamic task
    token = start_task(normalized_params)
    task_id = normalized_params["parameters"]["task_id"]
    task_info = wait_for_task_to_finish(token, task_id)
    task_params = __create_parameters(task_info)
    if task_info["done"]:
        raw_data = request_task_result(token)
        task_result = normalize_nodes(raw_data, parameters['identifier'])
        return Task(result=task_result, raw_data=raw_data, info=task_info, params=task_params)
    return Task(info=task_info, params=task_params)


def __create_parameters(info: dict) -> dict:
    algor = info["algorithm"]
    param = info["parameters"]
    param["algorithm"] = algor
    del param["inputNetwork"]
    return param


def __get_dict_for_genes_and_drugs(drugs: List[Drug], genes: List[Gene]) -> dict:
    r_drugs = {}
    r_genes = {}

    for d in drugs:
        r_drugs = {**r_drugs, **d.to_dict()}
    for g in genes:
        r_genes = {**r_genes, **g.to_dict()}

    return {"drugs": r_drugs, "genes": r_genes}



"""
drugstone.task.models.gene

This module implements the class Gene for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

from typing import List


class Gene:

    def __init__(self,
                 symbol: str = None,
                 protein_name: str = None,
                 hasEdgesTo: List[str] = None,
                 uniprot_ac: str = None,
                 entrez: str = None,
                 ensg: List[str] = None):
        self.__symbol = symbol
        self.__protein_name = protein_name
        self.__uniprot_ac = uniprot_ac
        self.__entrez = entrez
        self.__ensg = ensg
        self.__hasEdgesTo = hasEdgesTo or []

    def to_dict(self):
        return {
            self.__symbol: {
                "symbol": self.__symbol,
                "protein_name": self.__protein_name,
                "uniprot_ac": self.__uniprot_ac,
                "entrez": self.__entrez,
                "ensg": self.__ensg,
                "hasEdgesTo": self.__hasEdgesTo
            }
        }



import requests
import json
from .settings import API

def getAdjacentDrugs(normalized_params, extended_genes):
    url = f'{API}/adjacent_drugs/'
    seed_drugstone_ids = []
    for gene in extended_genes:
        seed_drugstone_ids.extend(gene['drugstoneId'])
    seed_drugstone_ids = [x[1:] for x in seed_drugstone_ids] # remove prefix p
    payload = {
        'proteins': seed_drugstone_ids,
        'pdi_dataset': normalized_params['parameters']['pdiDataset'],
        'licenced': normalized_params['parameters']['licenced']
    }
    payload = json.dumps(payload)
    headers = {'Content-Type': 'application/json; charset=UTF-8'}
    response = requests.post(url=url, data=payload, headers=headers)
    # format result into task result format
    result = response.json()
    result['genes'] = extended_genes
    return result




import string


def add_edges_to_genes(
        genes: list,
        edges: list, 
        identifier: string
        ) -> dict:

    for gene in genes:
        if "drugstoneId" in gene:
            netex_edges = [n["proteinB"] for n in edges if gene["drugstoneId"] == n["proteinA"]]
            symbol_edges = []
            for e in netex_edges:
                for g in genes:
                    if identifier in g and "drugstoneId" in g and e == g["drugstoneId"]:
                        symbol_edges.append(g[identifier])
            gene["hasEdgesTo"] = symbol_edges
        else:
            gene["hasEdgesTo"] = []

    result = {"drugs": {}, "genes": {}}

    for gene in genes:
        gene.pop("drugstoneId", None)
        result["genes"][gene["id"]] = gene

    return result



"""
drugstone.task.models.drug

This module implements the class Drug for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""


class Drug:

    def __init__(self,
                 label: str = None,
                 drug_id: str = None,
                 status: str = None):
        self.__label = label
        self.__drug_id = drug_id
        self.__status = status

    def to_dict(self):
        return {
            self.__label: {
                "label": self.__label,
                "drugId": self.__drug_id,
                "status": self.__status,
            }
        }



"""
drugstone.scripts.request_task_result

This module implements the request_task_result function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import requests
from .constants.url import Url


def request_task_result(token: str) -> dict:
    """Returns the raw task result."""

    url_parameter = "?view=&fmt=&token=" + token

    return requests.get(
        Url.TASK_RESULTS + url_parameter,
        verify=False
    ).json()



"""
drugstone.task.scripts.create_path

This module defines the create_path function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import os


def create_path(path: str, name: str, file_type: str) -> str:
    """Concatenates a path with a filename and a filetype.

    :param str path: A path.
    :param str name: A filename.
    :param str file_type: A filetype.
    :return: str: Resulting path.
    """
    f_type = "." + file_type
    n_path = os.path.join(path, name+f_type)
    path_exists = os.path.exists(n_path)
    iterations = 1
    while path_exists:
        if not n_path.endswith(")" + f_type):
            n_path = n_path[:len(n_path) - len(f_type)] + "(1)" + f_type
        else:
            n_path = n_path[:len(n_path) - len(f_type) - 2] + str(iterations) + ")" + f_type
            iterations += 1
        path_exists = os.path.exists(n_path)
    return n_path



"""
drugstone.scripts.normalize_nodes

This module implements the normalize_nodes function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""



def normalize_nodes(results: dict, identifier: str) -> dict:
    """Returns a normalized dict of the drugs and genes."""

    drugs = {}
    genes = {}
    drugstone_drug_id_to_label = {}

    # For all the nodes in the results,
    # differentiates if it's a drug or gene
    # and puts it in the according dict.
    nodes = results["nodeAttributes"]["details"]
    node_types = results["nodeAttributes"]["nodeTypes"]
    for n_id, n_type in node_types.items():
        nodes[n_id]["node_type"] = n_type
    for _, node in nodes.items():
        # init edge list for later
        node['hasEdgesTo'] = []
        node['isResult'] = False
        node['isConnector'] = False
        if node["node_type"] == "drug":
            n_name = node["label"]
            drugs[n_name] = node
            # needed because edge targets are drugstone IDs for drugs and not labels
            drugstone_drug_id_to_label[node['drugstoneId']] = n_name
        elif node["node_type"] == "protein":
            n_name = node[identifier][0]
            genes[n_name] = node

    # Adds to the genes if it's a seed or not.
    is_seed = results["nodeAttributes"]["isSeed"]
    for _, g_details in genes.items():
        g_details["is_seed"] = False if g_details[identifier][0] not in is_seed else is_seed[g_details["symbol"][0]]

    # Add information if node is result node
    for node in results['targetNodes']:
        try:
            # node is drug
            label = drugstone_drug_id_to_label[node]
            drugs[label]['isResult'] = True
        except KeyError:
            # node is gene
            genes[node]['isResult'] = True

    # add information if node is connector node
    if 'intermediateNodes' in results:
        for node in results['intermediateNodes']:
            genes[node]['isConnector'] = True

    # Normalizes the scores for the drugs.
    drug_scores = []
    for _, drug in drugs.items():
        if "score" in drug:
            drug_scores.append(drug["score"])
    if drug_scores:
        max_drug_score = max([x for x in drug_scores if x is not None])
    for _, drug in drugs.items():
        if "score" in drug and (type(drug["score"]) == int
                                or type(drug["score"]) == float):
            old_score = drug["score"]
            new_score = round(old_score / max_drug_score, 4)
            drug["score"] = new_score
        else:
            drug["score"] = None

    # sorts drugs without a score out
    none_score_drugs = []
    for drug, detail in drugs.items():
        if detail["score"] is None:
            none_score_drugs.append(drug)
    for x in none_score_drugs:
        drugs.pop(x)

    # Normalizes the scores for the genes.
    gene_scores = [1]
    for _, gene in genes.items():
        if "score" in gene:
            gene_scores.append(gene["score"])
    max_gene_score = max([x for x in gene_scores if x is not None])
    for _, gene in genes.items():
        if "score" in gene and (type(gene["score"]) == int
                                or type(gene["score"]) == float):
            old_score = gene["score"]
            new_score = round(old_score / max_gene_score, 4)
            gene["score"] = new_score
        else:
            gene["score"] = None

    # Adds the edges to the genes.
    edges = results["network"]["edges"]

    for e in edges:
        # test if it is a drug or gene edge
        if any([x in drugstone_drug_id_to_label for x in [e['from'], e['to']]]):
            # drug edge
            if e['from'] in drugstone_drug_id_to_label:
                label = drugstone_drug_id_to_label[e['from']]
                drugs[label]['hasEdgesTo'].append(e['to'])
            else:
                label = drugstone_drug_id_to_label[e['to']]
                drugs[label]['hasEdgesTo'].append(e['from'])
        else:
            # gene edge
            if any([x not in genes for x in [e['from'], e['to']]]):
                #  some edge targets are somehow not in the network, this should be fixed in the backend
                continue
            genes[e['from']]['hasEdgesTo'].append(e['to'])

    # Removes unnecessary properties from drugs.
    for _, drug in drugs.items():
        drug.pop("drugstoneId")
        drug.pop("trialLinks")
        drug.pop("node_type")

    # Removes unnecessary properties from genes.
    for _, gene in genes.items():
        gene.pop("drugstoneId")
        gene.pop("node_type")

    return {"drugs": drugs, "genes": genes}



"""
drugstone.task.scripts.download_network_graph

This module implements the download_network_graph function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

from pyvis.network import Network
from .create_path import create_path


def download_network_graph(nodes: dict, path: str = "", name: str = "graph") -> None:
    """Downloads a html file with a graph of the nodes."""

    # The network.
    net = Network(height='90%', width='100%')

    # Adds the nodes to the network.
    for g_name, gene in nodes["genes"].items():
        if "is_seed" in gene:
            if gene["is_seed"]:
                # net.add_node(g_name, g_name + " (seed)", color="blue", shape="dot")
                net.add_node(g_name, " ", color="blue", shape="dot")
            elif not gene["is_seed"]:
                # net.add_node(g_name, shape="triangle")
                net.add_node(g_name, " ", shape="triangle")
        else:
            net.add_node(g_name, g_name + " (static)", color="gray", shape="dot")
    for drug in nodes["drugs"]:
        net.add_node(drug, color="red", shape="square")

    # Adds the edges to the network.
    edges = []
    for n, gene in nodes["genes"].items():
        for e in gene["hasEdgesTo"]:
            edges.append((n, e))
    net.add_edges(edges)

    # net.show_buttons(filter_=['physics'])
    net.set_edge_smooth("cubicBezier")
    net.set_options("""
    var options = {
        "physics": {
            "enabled": true,
            "solver": "barnesHut",
            "barnesHut": {
                "gravitationalConstant": -250,
                "centralGravity": 0.001,
                "springLength": 50,
                "springConstant": 0.005,
                "avoidOverlap": 0.5
            },
            "stabilization": {
                "enabled": true,
                "iterations": 250
            }
        }
    }
    """)
    # net.toggle_physics(False)
    downloads_path = create_path(path, name, "html")
    net.show(downloads_path)

# private static analysisPhysics = {
#     enabled: true,
#     solver: 'barnesHut',
#     barnesHut: {
#       theta: 0.1,
#       gravitationalConstant: -50000,
#       centralGravity: 5,
#       springLength: 100,
#       springConstant: 0.8,
#       damping: 0.5,
#       avoidOverlap: 1,
#     },
#     stabilization: {
#       enabled: true,
#       iterations: 250
#     }
#   };

# "physics": {
#             "barnesHut": {
#             "gravitationalConstant": -250,
#             "centralGravity": 0,
#             "springLength": 50,
#             "springConstant": 0.005,
#             "avoidOverlap": 0.5
#             },
#             "minVelocity": 0.75
#         }



"""
drugstone.deep_search

This module implements the deep_search function for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

from typing import Union, List
from .task.task import Task
from .new_task import new_task


def deep_search(seeds: list, parameters: Union[dict, List[dict]] = dict({})) -> Task:
    """Performs a drug-target search of the seeds, followed by a drug search on the drug-target results.

    Starts with a drug-target search for the seed genes.
    Then uses the resulting targets to search for drugs.

    :param list seeds: List of seed nodes for the task.
    :param dict parameters: (optional) Dictionary of parameters for the task. Defaults to an empty dict {}.
    :return: :class:`Task` object
    """

    # task parameters
    t_params = {"algorithm": "keypathwayminer", "k": 10}
    d_params = {}

    if isinstance(parameters, list):
        if len(parameters) == 2:
            t_params = {**t_params, **parameters[0]}
            d_params = {**d_params, **parameters[1]}
        elif len(parameters) == 1:
            t_params = {**t_params, **parameters[0]}
            d_params = {**d_params, **parameters[0]}

    if isinstance(parameters, dict):
        for key, value in parameters.items():
            if isinstance(value, list) and len(value) > 1:
                t_params[key] = value[0]
                d_params[key] = value[1]
            else:
                t_params[key] = value
                d_params[key] = value

    t_params["target"] = "drug-target"
    d_params["target"] = "drug"
    if t_params.get("algorithm", None) == d_params.get("algorithm", None):
        t_params["has_duplicate_algorithms"] = True
        d_params["has_duplicate_algorithms"] = True

    # target search
    t_search = new_task(seeds, t_params)
    t_result = t_search.get_result().get_genes()
    # # keypathwayminer sometimes does not return all seeds
    targets = list({*seeds, *list(t_result.keys())})

    # drug search
    d_search = new_task(targets, d_params)
    return d_search



from . import settings
import requests
import logging

class license:
    printed = False
    accepted = False

license = license()

def print_license():
    license_text = requests.get(
        f'{settings.API}/get_license',
        verify=False).json()["license"]
    logging.info(license_text)
    global license
    license.printed = True
    return

def accept_license():
    global license
    if not license.printed:
        logging.error('license needs to be read first. Call "drugstone.print_license()".')
        return
    license.accepted = True
    logging.info('license accepted.')
    return



"""
drugstone

The python package for the https://drugst.one/ platform.
This package offers tools for drug-repurposing.
This is a programmatic approach to the functionality of the web portal.
For more information visit: https://drugst.one/

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import urllib3
import logging

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger().setLevel(logging.INFO)


from .new_task import new_task
from .new_tasks import new_tasks
from .deep_search import deep_search
from .task.models.drug import Drug
from .task.models.gene import Gene
from . import fetching
from .license import print_license, accept_license, license
